#!/usr/bin/env python3

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2019 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.3.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

import os
import sys
import uuid
import tempfile
import unittest
from datetime import datetime

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
PACKAGE_DIR = os.path.dirname(TEST_DIR)
sys.path.append(PACKAGE_DIR)

from anacore.illumina import Bcl2fastqLog, getInfFromSeqID, RTAComplete, RunInfo, RunParameters


########################################################################
#
# FUNCTIONS
#
########################################################################
class TestBcl2fastqLog(unittest.TestCase):
    def setUp(self):
        tmp_folder = tempfile.gettempdir()
        unique_id = str(uuid.uuid1())
        self.tmp_file = os.path.join(tmp_folder, unique_id + "_bcl2fastq_log.txt")
        self.test_cases = [
            {  # Complete
                "content": """BCL to FASTQ file converter
bcl2fastq v2.20.0.422
Copyright (c) 2007-2017 Illumina, Inc.

2021-06-03 17:18:10 [2afc55848400] Command-line invocation: bcl2fastq --no-lane-splitting --loading-threads 1 --writing-threads 1 --processing-threads 16 --runfolder-dir /NextSeq3/210602_NDX550500_RUO_0063_AHVW3TAFX2
2021-06-03 17:18:10 [2afc55848400] INFO: Minimum log level: INFO
2021-06-03 23:22:13 [2afc55848400] Processing completed with 0 errors and 1 warnings.""",
                "expected": {
                    "command": "bcl2fastq --no-lane-splitting --loading-threads 1 --writing-threads 1 --processing-threads 16 --runfolder-dir /NextSeq3/210602_NDX550500_RUO_0063_AHVW3TAFX2",
                    "complete": {
                        "nb_errors": 0,
                        "nb_warnings": 1
                    },
                    "version": "v2.20.0.422",
                    "start_time": datetime(2021, 6, 3, 17, 18, 10),
                    "end_time": datetime(2021, 6, 3, 23, 22, 13)
                }
            }, {  # Incomplete
                "content": """BCL to FASTQ file converter
bcl2fastq v2.20.0.422
Copyright (c) 2007-2017 Illumina, Inc.

2021-06-03 17:18:10 [2afc55848400] Command-line invocation: bcl2fastq --no-lane-splitting --loading-threads 1 --writing-threads 1 --processing-threads 16 --runfolder-dir /NextSeq3/210602_NDX550500_RUO_0063_AHVW3TAFX2
2021-06-03 17:18:10 [2afc55848400] INFO: Minimum log level: INFO""",
                "expected": {
                    "command": "bcl2fastq --no-lane-splitting --loading-threads 1 --writing-threads 1 --processing-threads 16 --runfolder-dir /NextSeq3/210602_NDX550500_RUO_0063_AHVW3TAFX2",
                    "complete": None,
                    "version": "v2.20.0.422",
                    "start_time": datetime(2021, 6, 3, 17, 18, 10),
                    "end_time": None
                }
            }
        ]

    def testParse(self):
        expected = [curr_test["expected"] for curr_test in self.test_cases]
        observed = []
        for curr_test in self.test_cases:
            with open(self.tmp_file, "w") as handle:
                handle.write(curr_test["content"] + "\n")
            res = Bcl2fastqLog(self.tmp_file)
            observed.append({
                "command": res.command,
                "complete": res.complete,
                "version": res.version,
                "start_time": res.start_time,
                "end_time": res.end_time
            })
        self.assertEqual(expected, observed)


class TestRTAComplete(unittest.TestCase):
    def setUp(self):
        tmp_folder = tempfile.gettempdir()
        unique_id = str(uuid.uuid1())
        self.tmp_file = os.path.join(tmp_folder, unique_id + "_RTAComplete.txt")
        self.test_cases = [
            {
                "content": "11/1/2017,15:11:43.174,Illumina RTA 1.18.54",
                "expected": {"RTA_version": "1.18.54", "end_date": datetime(2017, 11, 1, 15, 11, 43)}  # datetime.fromtimestamp(1484147503)
            }, {
                "content": "RTA 2.4.11 completed on 11/14/2019 4:56:45 AM",
                "expected": {"RTA_version": "2.4.11", "end_date": datetime(2019, 11, 14, 4, 56, 45)}  # datetime.fromtimestamp(1573707405)
            }, {
                "content": "RTA 2.4.11 completed on 11/14/2019 4:56:45 PM",
                "expected": {"RTA_version": "2.4.11", "end_date": datetime(2019, 11, 14, 16, 56, 45)}  # datetime.fromtimestamp(1573750605)
            }
        ]

    def testParse(self):
        expected = [curr_test["expected"] for curr_test in self.test_cases]
        observed = []
        for curr_test in self.test_cases:
            with open(self.tmp_file, "w") as handle:
                handle.write(curr_test["content"] + "\n")
            res = RTAComplete(self.tmp_file)
            observed.append({"RTA_version": res.RTA_version, "end_date": res.end_date})
        self.assertEqual(expected, observed)


class TestRunInfo(unittest.TestCase):
    def setUp(self):
        tmp_folder = tempfile.gettempdir()
        unique_id = str(uuid.uuid1())
        self.tmp_file = os.path.join(tmp_folder, unique_id + "_RunInfo.xml")
        self.test_cases = [
            {  # NexSeq dual index
                "content": '''<?xml version="1.0"?>
<RunInfo xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" Version="4">
  <Run Id="191113_NB551452_0101_AHW7JJAFXY" Number="101">
    <Flowcell>HW7JJAFXY</Flowcell>
    <Instrument>NB551452</Instrument>
    <Date>191113</Date>
    <Reads>
      <Read Number="1" NumCycles="71" IsIndexedRead="N" />
      <Read Number="2" NumCycles="17" IsIndexedRead="Y" />
      <Read Number="3" NumCycles="8" IsIndexedRead="Y" />
      <Read Number="4" NumCycles="71" IsIndexedRead="N" />
    </Reads>
    <FlowcellLayout LaneCount="4" SurfaceCount="2" SwathCount="1" TileCount="12" SectionPerLane="3" LanePerSection="2">
      <TileSet TileNamingConvention="FiveDigit">
        <Tiles>
          <Tile>1_11101</Tile>
        </Tiles>
      </TileSet>
    </FlowcellLayout>
    <ImageDimensions Width="2592" Height="1944" />
    <ImageChannels>
      <Name>Red</Name>
      <Name>Green</Name>
    </ImageChannels>
  </Run>
</RunInfo>''',
                "expected": {
                    "flowcell": {'id': 'HW7JJAFXY', 'layout': {'LaneCount': '4', 'SurfaceCount': '2', 'SwathCount': '1', 'TileCount': '12', 'SectionPerLane': '3', 'LanePerSection': '2'}},
                    "instrument": {'id': 'NB551452', 'platform': 'NextSeq'},
                    "reads_phases": [{'is_index': False, 'nb_cycles': 71}, {'is_index': True, 'nb_cycles': 17}, {'is_index': True, 'nb_cycles': 8}, {'is_index': False, 'nb_cycles': 71}],
                    "run": {'number': '101', 'id': '191113_NB551452_0101_AHW7JJAFXY', 'start_date': datetime(2019, 11, 13, 0, 0)}
                },
            },
            {  # NexSeq dual index
                "content": '''<?xml version="1.0"?>
<RunInfo xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" Version="4">
  <Run Id="200316_NDX550421_RUO_0003_AHJK57BGXF" Number="3">
    <Flowcell>HJK57BGXF</Flowcell>
    <Instrument>NDX550421_RUO</Instrument>
    <Date>200316</Date>
    <Reads>
      <Read Number="1" NumCycles="151" IsIndexedRead="N" />
      <Read Number="2" NumCycles="8" IsIndexedRead="Y" />
      <Read Number="3" NumCycles="8" IsIndexedRead="Y" />
      <Read Number="4" NumCycles="151" IsIndexedRead="N" />
    </Reads>
    <FlowcellLayout LaneCount="4" SurfaceCount="2" SwathCount="3" TileCount="12" SectionPerLane="3" LanePerSection="2">
      <TileSet TileNamingConvention="FiveDigit">
        <Tiles>
        </Tiles>
      </TileSet>
    </FlowcellLayout>
    <ImageDimensions Width="2592" Height="1944" />
    <ImageChannels>
      <Name>Red</Name>
      <Name>Green</Name>
    </ImageChannels>
  </Run>
</RunInfo>''',
                "expected": {
                    "flowcell": {'id': 'HJK57BGXF', 'layout': {'LaneCount': '4', 'SurfaceCount': '2', 'SwathCount': '3', 'TileCount': '12', 'SectionPerLane': '3', 'LanePerSection': '2'}},
                    "instrument": {'id': 'NDX550421_RUO', 'platform': 'NextSeq'},
                    "reads_phases": [{'is_index': False, 'nb_cycles': 151}, {'is_index': True, 'nb_cycles': 8}, {'is_index': True, 'nb_cycles': 8}, {'is_index': False, 'nb_cycles': 151}],
                    "run": {'number': '3', 'id': '200316_NDX550421_RUO_0003_AHJK57BGXF', 'start_date': datetime(2020, 3, 16, 0, 0)}
                },
            },
            {  # MiSeq dual index
                "content": '''<?xml version="1.0"?>
<RunInfo xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" Version="2">
  <Run Id="191120_M70265_0340_000000000-CRRBF" Number="340">
    <Flowcell>000000000-CRRBF</Flowcell>
    <Instrument>M70265</Instrument>
    <Date>191120</Date>
    <Reads>
      <Read NumCycles="151" Number="1" IsIndexedRead="N" />
      <Read NumCycles="8" Number="2" IsIndexedRead="Y" />
      <Read NumCycles="8" Number="3" IsIndexedRead="Y" />
      <Read NumCycles="151" Number="4" IsIndexedRead="N" />
    </Reads>
    <FlowcellLayout LaneCount="1" SurfaceCount="2" SwathCount="1" TileCount="19" />
  </Run>
</RunInfo>''',
                "expected": {
                    "flowcell": {'id': '000000000-CRRBF', 'layout': {'LaneCount': '1', 'SurfaceCount': '2', 'SwathCount': '1', 'TileCount': '19'}},
                    "instrument": {'id': 'M70265', 'platform': 'MiSeq'},
                    "reads_phases": [{'is_index': False, 'nb_cycles': 151}, {'is_index': True, 'nb_cycles': 8}, {'is_index': True, 'nb_cycles': 8}, {'is_index': False, 'nb_cycles': 151}],
                    "run": {'number': '340', 'id': '191120_M70265_0340_000000000-CRRBF', 'start_date': datetime(2019, 11, 20, 0, 0)}
                }
            },
            {  # HiSeq 2500 single index
                "content": '''<?xml version="1.0"?>
<RunInfo xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" Version="2">
  <Run Id="151014_D00154_0539_AHCTK2BCXX_HiSeq 2500_TruSeq Exome_12 replicates of NA12878" Number="539">
    <Flowcell>HCTK2BCXX</Flowcell>
    <Instrument>D00154</Instrument>
    <Date>151014</Date>
    <Reads>
      <Read Number="1" NumCycles="76" IsIndexedRead="N" />
      <Read Number="2" NumCycles="8" IsIndexedRead="Y" />
      <Read Number="3" NumCycles="76" IsIndexedRead="N" />
    </Reads>
    <FlowcellLayout LaneCount="2" SurfaceCount="2" SwathCount="2" TileCount="16" />
    <AlignToPhiX>
      <Lane>1</Lane>
      <Lane>2</Lane>
    </AlignToPhiX>
  </Run>
</RunInfo>''',
                "expected": {
                    "flowcell": {'id': 'HCTK2BCXX', 'layout': {'LaneCount': '2', 'SurfaceCount': '2', 'SwathCount': '2', 'TileCount': '16'}},
                    "instrument": {'id': 'D00154', 'platform': 'HiSeq'},
                    "reads_phases": [{'is_index': False, 'nb_cycles': 76}, {'is_index': True, 'nb_cycles': 8}, {'is_index': False, 'nb_cycles': 76}],
                    "run": {'number': '539', 'id': '151014_D00154_0539_AHCTK2BCXX_HiSeq 2500_TruSeq Exome_12 replicates of NA12878', 'start_date': datetime(2015, 10, 14, 0, 0)}
                }
            },
            {  # MiniSeq dual index
                "content": '''<?xml version="1.0"?>
<RunInfo xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" Version="4">
  <Run Id="160822_ML-P2-01_0042_A000H02MNK" Number="42">
    <Flowcell>000H02MNK</Flowcell>
    <Instrument>ML-P2-01</Instrument>
    <Date>160822</Date>
    <Reads>
      <Read Number="1" NumCycles="150" IsIndexedRead="N" />
      <Read Number="2" NumCycles="8" IsIndexedRead="Y" />
      <Read Number="3" NumCycles="8" IsIndexedRead="Y" />
      <Read Number="4" NumCycles="150" IsIndexedRead="N" />
    </Reads>
    <FlowcellLayout LaneCount="1" SurfaceCount="2" SwathCount="3" TileCount="10" SectionPerLane="1" LanePerSection="1">
      <TileSet TileNamingConvention="FiveDigit">
        <Tiles>
          <Tile>1_11102</Tile>
          <Tile>1_21102</Tile>
        </Tiles>
      </TileSet>
    </FlowcellLayout>
    <ImageDimensions Width="2592" Height="1944" />
    <ImageChannels>
      <Name>Red</Name>
      <Name>Green</Name>
    </ImageChannels>
  </Run>
</RunInfo>''',
                "expected": {
                    "flowcell": {'id': '000H02MNK', 'layout': {'LaneCount': '1', 'SurfaceCount': '2', 'SwathCount': '3', 'TileCount': '10', 'SectionPerLane': '1', 'LanePerSection': '1'}},
                    "instrument": {'id': 'ML-P2-01', 'platform': 'MiniSeq'},
                    "reads_phases": [{'is_index': False, 'nb_cycles': 150}, {'is_index': True, 'nb_cycles': 8}, {'is_index': True, 'nb_cycles': 8}, {'is_index': False, 'nb_cycles': 150}],
                    "run": {'number': '42', 'id': '160822_ML-P2-01_0042_A000H02MNK', 'start_date': datetime(2016, 8, 22, 0, 0)}
                }
            },
            {  # NovaSeq dual index from (https://github.com/broadinstitute/viral-ngs/tree/master/test/input/TestRunInfo)
                "content": '''<?xml version="1.0" encoding="utf-8"?>
<RunInfo>
    <Run Id="180627_SL-NVB_0098_AFCHCYTJDMXX" Number="98">
        <Flowcell>HCYTJDMXX</Flowcell>
        <Instrument>A00198</Instrument>
        <Date>6/27/2018 4:59:20 PM</Date>
        <Reads>
            <Read Number="1" NumCycles="101" IsIndexedRead="N"/>
            <Read Number="2" NumCycles="8" IsIndexedRead="Y"/>
            <Read Number="3" NumCycles="8" IsIndexedRead="Y"/>
            <Read Number="4" NumCycles="101" IsIndexedRead="N"/>
        </Reads>
        <FlowcellLayout LaneCount="2" SurfaceCount="2" SwathCount="4" TileCount="88" FlowcellSide="1">
            <TileSet TileNamingConvention="FourDigit">
                <Tiles>
                    <Tile>1_2101</Tile>
                </Tiles>
            </TileSet>
        </FlowcellLayout>
        <AlignToPhiX/>
        <ImageDimensions Width="3200" Height="3607"/>
        <ImageChannels>
            <Name>RED</Name>
            <Name>GREEN</Name>
        </ImageChannels>
    </Run>
</RunInfo>''',
                "expected": {
                    "flowcell": {'id': 'HCYTJDMXX', 'layout': {'LaneCount': '2', 'SurfaceCount': '2', 'SwathCount': '4', 'TileCount': '88', 'FlowcellSide': '1'}},
                    "instrument": {'id': 'A00198', 'platform': 'NovaSeq'},
                    "reads_phases": [{'is_index': False, 'nb_cycles': 101}, {'is_index': True, 'nb_cycles': 8}, {'is_index': True, 'nb_cycles': 8}, {'is_index': False, 'nb_cycles': 101}],
                    "run": {'number': '98', 'id': '180627_SL-NVB_0098_AFCHCYTJDMXX', 'start_date': datetime(2018, 6, 27, 16, 59, 20)}
                }
            }
        ]

    def testParse(self):
        expected = [curr_test["expected"] for curr_test in self.test_cases]
        observed = []
        for curr_test in self.test_cases:
            with open(self.tmp_file, "w") as handle:
                handle.write(curr_test["content"] + "\n")
            res = RunInfo(self.tmp_file)
            observed.append({
                "flowcell": res.flowcell,
                "instrument": res.instrument,
                "reads_phases": res.reads_phases,
                "run": res.run
            })
        self.assertEqual(expected, observed)


class TestRunParameters(unittest.TestCase):
    def setUp(self):
        tmp_folder = tempfile.gettempdir()
        unique_id = str(uuid.uuid1())
        self.tmp_file = os.path.join(tmp_folder, unique_id + "_RunParameters.xml")
        self.test_cases = [
            {  # NexSeq with post-process
                "content": '''<?xml version="1.0"?>
<RunParameters xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <Setup>
    <SupportMultipleSurfacesInUI>true</SupportMultipleSurfacesInUI>
    <ApplicationVersion>1.3.0.23</ApplicationVersion>
    <ApplicationName>NextSeq Control Software</ApplicationName>
    <NumTilesPerSwath>12</NumTilesPerSwath>
    <NumSwaths>3</NumSwaths>
    <NumLanes>4</NumLanes>
    <Read1>151</Read1>
    <Read2>151</Read2>
    <Index1Read>6</Index1Read>
    <Index2Read>0</Index2Read>
    <SectionPerLane>3</SectionPerLane>
    <LanePerSection>2</LanePerSection>
  </Setup>
  <RunID>141107_NS500413_0033_H14U5BGXX</RunID>
  <InstrumentID>NS500413</InstrumentID>
  <RunNumber>33</RunNumber>
  <RTAVersion>2.1.3</RTAVersion>
  <SystemSuiteVersion>1.3.0.9</SystemSuiteVersion>
  <FlowCellSerial>H14U5BGXX</FlowCellSerial>
  <PR2BottleSerial>NS2041161-BUFFR</PR2BottleSerial>
  <ReagentKitSerial>NS2044178-REAGT</ReagentKitSerial>
  <ReagentKitSerialWasEnteredInBaseSpace>false</ReagentKitSerialWasEnteredInBaseSpace>
  <ExperimentName>NEXTSEQ_HU01_CPCT_CPCT34WES</ExperimentName>
  <LibraryID>LibrariesPool_CPCT34WES</LibraryID>
  <Chemistry>NextSeq High</Chemistry>
  <SelectedTiles>
    <Tile>1_11101</Tile>
    <Tile>2_23612</Tile>
  </SelectedTiles>
  <RunFolder>D:\\Illumina\\NextSeq Control Software Temp\\141107_NS500413_0033_H14U5BGXX\\</RunFolder>
  <OutputFolder>\\NEXTSEQ500_HU01\\141107_NS500413_0033_H14U5BGXX\\</OutputFolder>
  <RecipeFolder>C:\\Illumina\\NextSeq Control Software\\Recipe\\High\\v1.1</RecipeFolder>
  <SimulationFolder />
  <RunStartDate>141107</RunStartDate>
  <FocusMethod>IXFocus</FocusMethod>
  <SurfaceToScan>Both</SurfaceToScan>
  <SaveFocusImages>false</SaveFocusImages>
  <SaveScanImages>false</SaveScanImages>
  <SelectiveSave>true</SelectiveSave>
  <IsPairedEnd>true</IsPairedEnd>
  <AnalysisWorkflowType>FastQ</AnalysisWorkflowType>
  <CustomReadOnePrimer>BP10</CustomReadOnePrimer>
  <CustomReadTwoPrimer>BP11</CustomReadTwoPrimer>
  <CustomIndexPrimer>BP12</CustomIndexPrimer>
  <CustomIndexTwoPrimer>EXT</CustomIndexTwoPrimer>
  <UsesCustomReadOnePrimer>false</UsesCustomReadOnePrimer>
  <UsesCustomReadTwoPrimer>false</UsesCustomReadTwoPrimer>
  <UsesCustomIndexPrimer>false</UsesCustomIndexPrimer>
  <UsesCustomIndexTwoPrimer>false</UsesCustomIndexTwoPrimer>
  <RunManagementType>BaseSpaceCloud</RunManagementType>
  <BaseSpaceRunId>9092087</BaseSpaceRunId>
  <BaseSpaceRunMode>UserAndIlluminaHealth</BaseSpaceRunMode>
  <ComputerName>NEXTSEQ500_HU01</ComputerName>
  <SequencingStarted>true</SequencingStarted>
  <PlannedRead1Cycles>151</PlannedRead1Cycles>
  <PlannedRead2Cycles>151</PlannedRead2Cycles>
  <PlannedIndex1ReadCycles>6</PlannedIndex1ReadCycles>
  <PlannedIndex2ReadCycles>0</PlannedIndex2ReadCycles>
  <IsRehyb>false</IsRehyb>
  <MaxCyclesSupportedByReagentKit>318</MaxCyclesSupportedByReagentKit>
</RunParameters>''',
                "expected": {
                    "instrument": {"id": "NS500413", "platform": "NextSeq"},
                    "reads_phases": [{'is_index': False, 'nb_cycles': 151}, {'is_index': True, 'nb_cycles': 6}, {'is_index': True, 'nb_cycles': 0}, {'is_index': False, 'nb_cycles': 151}],
                    "run": {"number": "33", "id": "141107_NS500413_0033_H14U5BGXX", "start_date": datetime(2014, 11, 7)},
                    "kit": {"flowcell_id": "H14U5BGXX", "reagent_kit_id": "NS2044178-REAGT"},
                    "post_process": "FastQ",
                    "software": {"RTA": "2.1.3", "CS": "1.3.0.23"}
                }
            },
            {  # NexSeq without post-process
                "content": '''<?xml version="1.0"?>
<RunParameters xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <RunParametersVersion>NextSeq_4_0_0</RunParametersVersion>
  <Setup>
    <SupportMultipleSurfacesInUI>true</SupportMultipleSurfacesInUI>
    <ApplicationVersion>4.0.1.41</ApplicationVersion>
    <ApplicationName>NextSeq Control Software</ApplicationName>
    <NumTilesPerSwath>12</NumTilesPerSwath>
    <NumSwaths>1</NumSwaths>
    <NumLanes>4</NumLanes>
    <Read1>151</Read1>
    <Read2>151</Read2>
    <Index1Read>8</Index1Read>
    <Index2Read>8</Index2Read>
    <SectionPerLane>3</SectionPerLane>
    <LanePerSection>2</LanePerSection>
  </Setup>
  <RunID>200318_NDX550421_RUO_0004_AH5THYAFX2</RunID>
  <CopyServiceRunId>16A9530EA81D22F4</CopyServiceRunId>
  <InstrumentID>NDX550421_RUO</InstrumentID>
  <RunNumber>4</RunNumber>
  <RTAVersion>2.11.3</RTAVersion>
  <SystemSuiteVersion>4.0.1.41</SystemSuiteVersion>
  <LocalRunManagerVersion>2.4.0.2385</LocalRunManagerVersion>
  <RecipeVersion>4.0.0.1</RecipeVersion>
  <FirmwareVersion>r.2019.2.0</FirmwareVersion>
  <FlowCellRfidTag>
    <SerialNumber>H5THYAFX2</SerialNumber>
    <PartNumber>20022409</PartNumber>
    <LotNumber>20428008</LotNumber>
    <ExpirationDate>2022-02-04T00:00:00</ExpirationDate>
  </FlowCellRfidTag>
  <PR2BottleRfidTag>
    <SerialNumber>NS4071554-BUFFR</SerialNumber>
    <PartNumber>15057941</PartNumber>
    <LotNumber>20429662</LotNumber>
    <ExpirationDate>2021-02-09T00:00:00</ExpirationDate>
  </PR2BottleRfidTag>
  <ReagentKitRfidTag>
    <SerialNumber>NS4065291-REAGT</SerialNumber>
    <PartNumber>15057939</PartNumber>
    <LotNumber>20426104</LotNumber>
    <ExpirationDate>2021-02-04T00:00:00</ExpirationDate>
  </ReagentKitRfidTag>
  <FlowCellSerial>H5THYAFX2</FlowCellSerial>
  <PR2BottleSerial>NS4071554-BUFFR</PR2BottleSerial>
  <ReagentKitSerial>NS4065291-REAGT</ReagentKitSerial>
  <ReagentKitSerialWasEnteredInBaseSpace>false</ReagentKitSerialWasEnteredInBaseSpace>
  <ExperimentName>200318</ExperimentName>
  <LibraryID />
  <StateDescription />
  <Chemistry>NextSeq Mid</Chemistry>
  <ChemistryVersion>2.1</ChemistryVersion>
  <SelectedTiles>
    <Tile>1_11101</Tile>
    <Tile>4_21612</Tile>
  </SelectedTiles>
  <RunFolder>D:\\Illumina\\NextSeq Control Software Temp\\200318_NDX550421_RUO_0004_AH5THYAFX2\\</RunFolder>
  <RTALogsFolder />
  <PreRunFolderRoot>D:\\Illumina\\NextSeq Control Software PreRun</PreRunFolderRoot>
  <PreRunFolder>D:\\Illumina\\NextSeq Control Software PreRun\\NDX550421_RUO_2020-03-18__16_36_39</PreRunFolder>
  <OutputFolder>\\NextSeq3\\200318_NDX550421_RUO_0004_AH5THYAFX2\\</OutputFolder>
  <RecipeFolder>C:\\Program Files\\Illumina\\NextSeq Control Software\\Recipe\\Mid\\v2.5</RecipeFolder>
  <SimulationFolder />
  <RunStartDate>200318</RunStartDate>
  <BaseSpaceUserName />
  <LocalRunManagerUserName />
  <FocusMethod>IXFocus</FocusMethod>
  <SurfaceToScan>Both</SurfaceToScan>
  <SaveFocusImages>false</SaveFocusImages>
  <SaveScanImages>false</SaveScanImages>
  <SelectiveSave>true</SelectiveSave>
  <IsPairedEnd>true</IsPairedEnd>
  <AnalysisWorkflowType />
  <CustomReadOnePrimer>BP10</CustomReadOnePrimer>
  <CustomReadTwoPrimer>BP11</CustomReadTwoPrimer>
  <CustomIndexOnePrimer>BP14</CustomIndexOnePrimer>
  <CustomIndexTwoPrimer>BP14</CustomIndexTwoPrimer>
  <UsesCustomReadOnePrimer>false</UsesCustomReadOnePrimer>
  <UsesCustomReadTwoPrimer>false</UsesCustomReadTwoPrimer>
  <UsesCustomIndexPrimer>false</UsesCustomIndexPrimer>
  <UsesCustomIndexTwoPrimer>false</UsesCustomIndexTwoPrimer>
  <BaseSpaceRunId>194270150</BaseSpaceRunId>
  <LocalRunManagerRunId xsi:nil="true" />
  <RunSetupType>ManualRunSetup</RunSetupType>
  <RunMode>PerformanceDataOnly</RunMode>
  <ComputerName>NDX550421-RUO</ComputerName>
  <SequencingStarted>true</SequencingStarted>
  <PlannedRead1Cycles>151</PlannedRead1Cycles>
  <PlannedRead2Cycles>151</PlannedRead2Cycles>
  <PlannedIndex1ReadCycles>8</PlannedIndex1ReadCycles>
  <PlannedIndex2ReadCycles>8</PlannedIndex2ReadCycles>
  <IsRehyb>false</IsRehyb>
  <PurgeConsumables>true</PurgeConsumables>
  <MaxCyclesSupportedByReagentKit>318</MaxCyclesSupportedByReagentKit>
  <ModuleName />
  <ModuleVersion />
</RunParameters>''',
                "expected": {
                    "instrument": {"id": "NDX550421_RUO", "platform": "NextSeq"},
                    "reads_phases": [{'is_index': False, 'nb_cycles': 151}, {'is_index': True, 'nb_cycles': 8}, {'is_index': True, 'nb_cycles': 8}, {'is_index': False, 'nb_cycles': 151}],
                    "run": {"number": "4", "id": "200318_NDX550421_RUO_0004_AH5THYAFX2", "start_date": datetime(2020, 3, 18)},
                    "kit": {"flowcell_id": "H5THYAFX2", "reagent_kit_id": "NS4065291-REAGT"},
                    "post_process": None,
                    "software": {"RTA": "2.11.3", "CS": "4.0.1.41"}
                }
            },
            {  # MiSeq
                "content": '''<?xml version="1.0"?>
<RunParameters xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <EnableCloud>true</EnableCloud>
  <RunParametersVersion>MiSeq_1_1</RunParametersVersion>
  <CopyManifests>true</CopyManifests>
  <FlowcellRFIDTag>
    <SerialNumber>000000000-AAC1Y</SerialNumber>
    <PartNumber>15028382</PartNumber>
    <ExpirationDate>2015-05-28T00:00:00</ExpirationDate>
  </FlowcellRFIDTag>
  <PR2BottleRFIDTag>
    <SerialNumber>MS2313945-00PR2</SerialNumber>
    <PartNumber>15041807</PartNumber>
    <ExpirationDate>2015-05-18T00:00:00</ExpirationDate>
  </PR2BottleRFIDTag>
  <ReagentKitRFIDTag>
    <SerialNumber>MS2740035-300V2</SerialNumber>
    <PartNumber>15033572</PartNumber>
    <ExpirationDate>2015-06-12T00:00:00</ExpirationDate>
  </ReagentKitRFIDTag>
  <Resumable>true</Resumable>
  <ManifestFiles />
  <AfterRunWashMethod>Post-Run Wash</AfterRunWashMethod>
  <Setup>
    <SupportMultipleSurfacesInUI>true</SupportMultipleSurfacesInUI>
    <ApplicationVersion>2.5.0.5</ApplicationVersion>
    <NumTilesPerSwath>14</NumTilesPerSwath>
    <NumSwaths>1</NumSwaths>
    <NumLanes>1</NumLanes>
    <ApplicationName>MiSeq Control Software</ApplicationName>
  </Setup>
  <RunID>141105_M01102_0149_000000000-AAC1Y</RunID>
  <ScannerID>M01102</ScannerID>
  <RunNumber>149</RunNumber>
  <FPGAVersion>9.5.12</FPGAVersion>
  <MCSVersion>2.5.0.5</MCSVersion>
  <RTAVersion>1.18.54</RTAVersion>
  <Barcode>000000000-AAC1Y</Barcode>
  <PR2BottleBarcode>MS2313945-00PR2</PR2BottleBarcode>
  <ReagentKitPartNumberEntered>15033572</ReagentKitPartNumberEntered>
  <ReagentKitVersion>Version2</ReagentKitVersion>
  <ReagentKitBarcode>MS2740035-300V2</ReagentKitBarcode>
  <PreviousReagentKitBarcode>MS2740035-300V2</PreviousReagentKitBarcode>
  <ExperimentName>MISEQ_UMC01_CVERMEULEN_WDL004</ExperimentName>
  <Chemistry>Amplicon</Chemistry>
  <Username>sbsuser</Username>
  <Workflow>
    <Analysis>GenerateFASTQ</Analysis>
  </Workflow>
  <EnableAnalysis>false</EnableAnalysis>
  <Reads>
    <RunInfoRead Number="1" NumCycles="151" IsIndexedRead="N" />
    <RunInfoRead Number="2" NumCycles="8" IsIndexedRead="Y" />
    <RunInfoRead Number="3" NumCycles="8" IsIndexedRead="Y" />
    <RunInfoRead Number="4" NumCycles="151" IsIndexedRead="N" />
  </Reads>
  <TempFolder>D:\\Illumina\\MiSeqTemp\\141105_M01102_0149_000000000-AAC1Y</TempFolder>
  <AnalysisFolder>D:\\Illumina\\MiSeqAnalysis\\141105_M01102_0149_000000000-AAC1Y</AnalysisFolder>
  <RunStartDate>141105</RunStartDate>
  <MostRecentWashType>PostRun</MostRecentWashType>
  <RecipeFolder>D:\\Illumina\\MiSeq Control Software\\CustomRecipe</RecipeFolder>
  <ILMNOnlyRecipeFolder>C:\\Illumina\MiSeq Control Software\\Recipe</ILMNOnlyRecipeFolder>
  <SampleSheetName>MISEQ_UMC01_CVERMEULEN_WDL004</SampleSheetName>
  <SampleSheetFolder>C:\\Users\\sbsuser\\AppData\\Roaming\\Illumina\\Illumina Experiment Manager\\Sample Sheets</SampleSheetFolder>
  <ManifestFolder>D:\\Illumina\\MiSeq Control Software\\Manifests</ManifestFolder>
  <OutputFolder>\\results_miseq_umc01\\141105_M01102_0149_000000000-AAC1Y</OutputFolder>
  <FocusMethod>AutoFocus</FocusMethod>
  <SurfaceToScan>Both</SurfaceToScan>
  <SaveFocusImages>true</SaveFocusImages>
  <SaveScanImages>true</SaveScanImages>
  <CloudUsername>user@institute.eu</CloudUsername>
  <RunManagementType>Standalone</RunManagementType>
  <CloudRunId>9058072</CloudRunId>
  <SendInstrumentHealthToILMN>false</SendInstrumentHealthToILMN>
</RunParameters>''',
                "expected": {
                    "instrument": {"id": "M01102", "platform": "MiSeq"},
                    "reads_phases": [{'is_index': False, 'nb_cycles': 151}, {'is_index': True, 'nb_cycles': 8}, {'is_index': True, 'nb_cycles': 8}, {'is_index': False, 'nb_cycles': 151}],
                    "run": {"number": "149", "id": "141105_M01102_0149_000000000-AAC1Y", "start_date": datetime(2014, 11, 5)},
                    "kit": {"flowcell_id": "000000000-AAC1Y", "reagent_kit_id": "MS2740035-300V2"},
                    "post_process": "GenerateFASTQ",
                    "software": {"RTA": "1.18.54", "CS": "2.5.0.5"}
                }
            },
            {  # HiSeq
                "content": '''<?xml version="1.0"?>
<RunParameters xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <Setup>
    <ExperimentName>HiSeq_UMC01_B_Diag_HAMVRADXX</ExperimentName>
    <ScanID>D00267</ScanID>
    <FCPosition>B</FCPosition>
    <WorkFlowType>SINGLEINDEX</WorkFlowType>
    <PairEndFC>true</PairEndFC>
    <Read1>101</Read1>
    <IndexRead1>7</IndexRead1>
    <IndexRead2>0</IndexRead2>
    <Read2>101</Read2>
    <OutputFolder>Z:\\</OutputFolder>
    <CompressBcls>true</CompressBcls>
    <RemapQScores />
    <NearNeighborCorrection>false</NearNeighborCorrection>
    <PeriodicSave>Save All Thumbnails</PeriodicSave>
    <Flowcell>HiSeq Rapid Flow Cell v1</Flowcell>
    <FirstBaseConfirmation>false</FirstBaseConfirmation>
    <SampleSheet>C:\\Users\\sbsuser\\Desktop\\sample sheets\\Diagnostic Sample Sheets\\HiSeq_UMC01_B_Diag_HAMVRADXX.csv</SampleSheet>
    <KeepIntensityFiles>false</KeepIntensityFiles>
    <Sbs>TruSeq Rapid SBS Kit</Sbs>
    <Pe>TruSeq Rapid PE Cluster Kit</Pe>
    <Index>TruSeq Rapid PE Cluster Kit</Index>
    <AlignToPhiX>
      <Lane>1</Lane>
      <Lane>2</Lane>
    </AlignToPhiX>
    <ClusteringChoice>OnBoardClustering</ClusteringChoice>
    <RunMode>RapidRun</RunMode>
    <Rehyb>None</Rehyb>
    <PerformPreRunFluidicsCheck>false</PerformPreRunFluidicsCheck>
    <ServiceRun>false</ServiceRun>
    <ApplicationName>HiSeq Control Software</ApplicationName>
    <ApplicationVersion>2.2.38</ApplicationVersion>
    <RunID>141110_D00267_0193_BHAMVRADXX</RunID>
    <RunStartDate>141110</RunStartDate>
    <IntegrationMode>BaseSpace</IntegrationMode>
    <BaseSpaceSettings>
      <RunMonitoringOnly>true</RunMonitoringOnly>
      <Username>user@institute.nl</Username>
      <PlannedRun>false</PlannedRun>
      <RunId>9136142</RunId>
      <TempFolder>E:\\Illumina\\BaseSpaceTemp</TempFolder>
      <SendInstrumentHealthToILMN>true</SendInstrumentHealthToILMN>
    </BaseSpaceSettings>
    <ScannerID>D00267</ScannerID>
    <ScanNumber>193</ScanNumber>
    <ComputerName>HWI-D00267</ComputerName>
    <FPGAVersion>7.9.7</FPGAVersion>
    <CPLDVersion>3.0.0</CPLDVersion>
    <RTAVersion>1.18.61</RTAVersion>
    <ChemistryVersion>Illumina,Bruno Fluidics Controller,0,v2.0340</ChemistryVersion>
    <CameraFirmware>2.01-F20-R02</CameraFirmware>
    <CameraDriver>6.45.20.3690</CameraDriver>
    <FocusCameraFirmware />
    <Barcode>HAMVRADXX</Barcode>
    <Username>sbsuser</Username>
    <SelectedSections>
      <Section Name="A_1" />
      <Section Name="B_1" />
    </SelectedSections>
    <FocusMethod>DynamicITF</FocusMethod>
    <SelectedSurface>BothLaneSurfaces</SelectedSurface>
    <SwathScanMode>DualSwathFC</SwathScanMode>
    <EnableLft>true</EnableLft>
    <AutoTiltOnce>true</AutoTiltOnce>
    <EnableAutoCenter>true</EnableAutoCenter>
    <EnableAnalysis>true</EnableAnalysis>
    <EnableBasecalling>true</EnableBasecalling>
    <EnableCameraLogging>false</EnableCameraLogging>
    <AdapterPlate>HiSeq Adapter Plate</AdapterPlate>
    <SlideHolder>HiSeq Flow Cell Holder</SlideHolder>
    <TemplateCycleCount>4</TemplateCycleCount>
    <NumAnalysisThreads>8</NumAnalysisThreads>
    <FPGADynamicFocusSettings>
      <MaxInitialZJumpHalfUm>3</MaxInitialZJumpHalfUm>
      <MaxSubsequentZJumpHalfUm>7</MaxSubsequentZJumpHalfUm>
      <NumberOfInitialZJumps>0</NumberOfInitialZJumps>
      <CVGainStart>500</CVGainStart>
      <CVGainPosLocked>500</CVGainPosLocked>
      <Offset>250</Offset>
      <HotPixel>350</HotPixel>
      <MotorDelayFrames>10</MotorDelayFrames>
      <SoftwareLaserLag>200</SoftwareLaserLag>
      <DitherSize>100</DitherSize>
      <GroupSize>50</GroupSize>
      <DitherShift>0</DitherShift>
      <IntensityCeiling>65535</IntensityCeiling>
      <IGain>100</IGain>
      <IHistory>4</IHistory>
    </FPGADynamicFocusSettings>
    <TileWidth>2048</TileWidth>
    <TileHeight>10000</TileHeight>
    <ImageWidth>2048</ImageWidth>
    <ImageHeight>160000</ImageHeight>
    <AreaPerPixelmm2>1.40625E-07</AreaPerPixelmm2>
    <LaneLength>60</LaneLength>
    <NumTilesPerSwath>16</NumTilesPerSwath>
    <NumSwaths>2</NumSwaths>
    <UseExistingRecipe>false</UseExistingRecipe>
    <Reads>
      <Read Number="1" NumCycles="101" IsIndexedRead="N" />
      <Read Number="2" NumCycles="7" IsIndexedRead="Y" />
      <Read Number="3" NumCycles="101" IsIndexedRead="N" />
    </Reads>
    <EnableNotifications>false</EnableNotifications>
    <ReagentKits>
      <Sbs>
        <SbsReagentKit>
          <ID>9806021</ID>
          <Prime>false</Prime>
          <NumberCyclesRemaining>225</NumberCyclesRemaining>
          <IsNew50Cycle>false</IsNew50Cycle>
          <IsNew200Cycle>true</IsNew200Cycle>
        </SbsReagentKit>
      </Sbs>
      <Index>
        <ReagentKit>
          <ID>9991543</ID>
        </ReagentKit>
      </Index>
      <Pe />
      <Rehyb />
    </ReagentKits>
    <ReagentBottles>
      <Sbs />
    </ReagentBottles>
    <Resume>false</Resume>
    <ResumeCycle>0</ResumeCycle>
    <SupportMultipleSurfacesInUI>true</SupportMultipleSurfacesInUI>
    <TempFolder>E:\\Illumina\\HiSeqTemp\\141110_D00267_0193_BHAMVRADXX</TempFolder>
    <RecipeFragmentVersion>1.5.14.0</RecipeFragmentVersion>
    <PromptForPeReagents>false</PromptForPeReagents>
    <MockRun>false</MockRun>
    <ScannedBarcode />
  </Setup>
  <Version>1</Version>
</RunParameters>''',
                "expected": {
                    "instrument": {"id": "D00267", "platform": "HiSeq"},
                    "reads_phases": [{'is_index': False, 'nb_cycles': 101}, {'is_index': True, 'nb_cycles': 7}, {'is_index': False, 'nb_cycles': 101}],
                    "run": {"number": "193", "id": "141110_D00267_0193_BHAMVRADXX", "start_date": datetime(2014, 11, 10)},
                    "kit": {"flowcell_id": "HAMVRADXX", "reagent_kit_id": None},
                    "post_process": None,
                    "software": {"RTA": "1.18.61", "CS": "2.2.38"}
                }
            }
        ]

    def testParse(self):
        expected = [curr_test["expected"] for curr_test in self.test_cases]
        observed = []
        for curr_test in self.test_cases:
            with open(self.tmp_file, "w") as handle:
                handle.write(curr_test["content"] + "\n")
            res = RunParameters(self.tmp_file)
            observed.append({
                "instrument": res.instrument,
                "reads_phases": res.reads_phases,
                "run": res.run,
                "kit": res.kit,
                "post_process": res.post_process,
                "software": res.software
            })
        self.assertEqual(expected, observed)


class TestFunctions(unittest.TestCase):
    def testGetInfFromSeqID(self):
        # Whithout UMI
        expected = {
            "sequencer_id": "NDX550421_RUO",
            "run_id": "15",
            "flowcell_id": "HY53NBGXC",
            "lane_id": 1,
            "tile_id": 11101,
            "x_pos": 9518,
            "y_pos": 1037,
            "umi": None
        }
        observed = getInfFromSeqID("NDX550421_RUO:15:HY53NBGXC:1:11101:9518:1037")
        self.assertEqual(expected, observed)
        # With UMI
        expected = {
            "sequencer_id": "NDX550421_RUO",
            "run_id": "15",
            "flowcell_id": "HY53NBGXC",
            "lane_id": 1,
            "tile_id": 11101,
            "x_pos": 9518,
            "y_pos": 1037,
            "umi": "TTGCCNG+CCTCATA"
        }
        observed = getInfFromSeqID("NDX550421_RUO:15:HY53NBGXC:1:11101:9518:1037:TTGCCNG+CCTCATA")
        self.assertEqual(expected, observed)


########################################################################
#
# MAIN
#
########################################################################
if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
#
# Copyright (C) 2018 IUCT-O
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2018 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.0.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

import os
import sys
import uuid
import tempfile
import unittest

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
PACKAGE_DIR = os.path.dirname(TEST_DIR)
sys.path.append(PACKAGE_DIR)

from anacore.gff import GFF3IO, GFF3Record


########################################################################
#
# FUNCTIONS
#
########################################################################
def diffRecord(record_a, record_b):
    are_diff = False
    for attr_name in ["seq_id", "source", "type", "start", "end", "score", "strand", "phase"]:
        if getattr(record_a, attr_name) != getattr(record_b, attr_name):
            are_diff = True
    if sorted(record_a.annot.keys()) != sorted(record_b.annot.keys()):
        are_diff = True
    else:
        for key in sorted(record_a.annot.keys()):
            if record_a.annot[key] != record_b.annot[key]:
                are_diff = True
    return are_diff


class TestGTFIO(unittest.TestCase):
    def setUp(self):
        tmp_folder = tempfile.gettempdir()
        unique_id = str(uuid.uuid1())

        # Temporary files
        self.tmp_in_gff = os.path.join(tmp_folder, unique_id + "_in.gff3")
        self.tmp_out_gff = os.path.join(tmp_folder, unique_id + "_out.gff3")

        self.expected_records = [
            GFF3Record("ctg123", None, "operon", 1300, 15000, None, "+", None, {"ID": "operon001", "Name": "superOperon"}),
            GFF3Record("ctg123", None, "mRNA", 1300, 9000, None, "+", None, {"ID": "mrna0001", "Name": "sonichedgehog", "Parent": "operon001"}),
            GFF3Record("ctg123", None, "exon", 1300, 1500, None, "+", None, {"Parent": "mrna0001"}),
            GFF3Record("ctg123", None, "exon", 1050, 1500, None, "+", None, {"Parent": "mrna0001"}),
            GFF3Record("ctg123", None, "exon", 3000, 3902, None, "+", None, {"Parent": "mrna0001"}),
            GFF3Record("ctg123", None, "exon", 5000, 5500, None, "+", None, {"Parent": "mrna0001"}),
            GFF3Record("ctg123", None, "exon", 7000, 9000, None, "+", None, {"Parent": "mrna0001"}),
            GFF3Record("ctg123", None, "mRNA", 10000, 15000, None, "+", None, {"ID": "mrna0002", "Name": "subsonicsquirrel", "Parent": "operon001", "note": "test	encode and decode (cleanning characters ;, =)."}),
            GFF3Record("ctg123", None, "exon", 10000, 12000, None, "+", None, {"Parent": "mrna0002"}),
            GFF3Record("ctg123", None, "exon", 14000, 15000, None, "+", None, {"Parent": "mrna0002"})
        ]
        self.expected_content = """##gff-version 3
ctg123	.	operon	1300	15000	.	+	.	Name=superOperon;ID=operon001
ctg123	.	mRNA	1300	9000	.	+	.	Name=sonichedgehog;ID=mrna0001;Parent=operon001
ctg123	.	exon	1300	1500	.	+	.	Parent=mrna0001
ctg123	.	exon	1050	1500	.	+	.	Parent=mrna0001
ctg123	.	exon	3000	3902	.	+	.	Parent=mrna0001
ctg123	.	exon	5000	5500	.	+	.	Parent=mrna0001
ctg123	.	exon	7000	9000	.	+	.	Parent=mrna0001
ctg123	.	mRNA	10000	15000	.	+	.	Name=subsonicsquirrel;ID=mrna0002;Parent=operon001;note=test	encode and decode (cleanning characters %3B%2C %3D).
ctg123	.	exon	10000	12000	.	+	.	Parent=mrna0002
ctg123	.	exon	14000	15000	.	+	.	Parent=mrna0002"""
        with open(self.tmp_in_gff, "w") as FH_gtf:
            FH_gtf.write(self.expected_content)

    def testWrite(self):
        # Write
        with GFF3IO(self.tmp_out_gff, "w") as FH_out:
            for record in self.expected_records:
                FH_out.write(record)
        # Read
        observed_content = None
        with open(self.tmp_out_gff) as FH_out:
            observed_content = "".join(FH_out.readlines()).strip()
        # Assert
        self.assertEqual(
            self.expected_content.replace("test	encode and decode", "test encode and decode"),
            observed_content
        )

    def testRead(self):
        # Read
        observed_records = []
        with GFF3IO(self.tmp_in_gff) as FH_in:
            for record in FH_in:
                observed_records.append(record)
        # Assert
        self.assertTrue(len(self.expected_records) == len(observed_records))
        for record_a, record_b in zip(self.expected_records, observed_records):
            self.assertFalse(diffRecord(record_a, record_b))

    def tearDown(self):
        # Clean temporary files
        for curr_file in [self.tmp_in_gff, self.tmp_out_gff]:
            if os.path.exists(curr_file):
                os.remove(curr_file)


########################################################################
#
# MAIN
#
########################################################################
if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2018 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.2.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

import os
import sys
import unittest

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
PACKAGE_DIR = os.path.dirname(TEST_DIR)
sys.path.append(PACKAGE_DIR)

from anacore.region import Region, RegionTree, RegionList, splittedByRef, iterOverlapped, consolidated


########################################################################
#
# FUNCTIONS
#
########################################################################
class TestRegion(unittest.TestCase):
    def testLength(self):
        self.assertEqual(Region(9, 15, None).length(), 7)
        self.assertEqual(Region(9, 15, "+").length(), 7)
        self.assertEqual(Region(9, 15, "-").length(), 7)
        self.assertEqual(Region(9, None, "-").length(), 1)
        self.assertEqual(Region(9, 9, "-").length(), 1)

    def testGetPosOnRef(self):
        # Forward strand
        f_stranded_region = Region(9, 15, "+")
        self.assertEqual(f_stranded_region.getPosOnRef(1), 9)  # First nt
        self.assertEqual(f_stranded_region.getPosOnRef(3), 11)
        self.assertEqual(f_stranded_region.getPosOnRef(7), 15)  # Last nt
        # self.assertRaises(ValueError, f_stranded_region.getPosOnRef(8))  # Out of region
        # Reverse strand
        r_stranded_region = Region(9, 15, "-")
        self.assertEqual(r_stranded_region.getPosOnRef(1), 15)  # First nt
        self.assertEqual(r_stranded_region.getPosOnRef(3), 13)
        self.assertEqual(r_stranded_region.getPosOnRef(7), 9)  # Last nt
        # self.assertRaises(ValueError, r_stranded_region.getPosOnRef(8))  # Out of region

    def testGetPosOnRegion(self):
        # Forward strand
        f_stranded_region = Region(9, 15, "+", "chr1")
        self.assertEqual(f_stranded_region.getPosOnRegion(9), 1)  # First nt
        self.assertEqual(f_stranded_region.getPosOnRegion(11), 3)
        self.assertEqual(f_stranded_region.getPosOnRegion(15), 7)  # Last nt
        with self.assertRaises(ValueError):
            f_stranded_region.getPosOnRegion(8)  # Out of region
        with self.assertRaises(ValueError):
            f_stranded_region.getPosOnRegion(16)  # Out of region
        # Reverse strand
        r_stranded_region = Region(9, 15, "-", "chr1")
        self.assertEqual(r_stranded_region.getPosOnRegion(15), 1)  # First nt
        self.assertEqual(r_stranded_region.getPosOnRegion(13), 3)
        self.assertEqual(r_stranded_region.getPosOnRegion(9), 7)  # Last nt
        with self.assertRaises(ValueError):
            r_stranded_region.getPosOnRegion(8)  # Out of region
        with self.assertRaises(ValueError):
            r_stranded_region.getPosOnRegion(16)  # Out of region

    def testContains(self):
        container_region = Region(9, 15, "+", "chr1")
        self.assertEqual(
            container_region.contains(Region(9, 9, "+", "chr1")),
            True
        )
        self.assertEqual(
            container_region.contains(Region(15, 15, "+", "chr1")),
            True
        )
        self.assertEqual(
            container_region.contains(Region(12, 13, "+", "chr1")),
            True
        )
        self.assertEqual(
            container_region.contains(Region(9, 15, "+", "chr1")),
            True
        )
        self.assertEqual(
            container_region.contains(Region(8, 14, "+", "chr1")),
            False
        )
        self.assertEqual(
            container_region.contains(Region(10, 16, "+", "chr1")),
            False
        )
        self.assertEqual(
            container_region.contains(Region(8, 16, "+", "chr1")),
            False
        )
        self.assertEqual(
            container_region.contains(Region(8, 8, "+", "chr1")),
            False
        )
        self.assertEqual(
            container_region.contains(Region(16, 16, "+", "chr1")),
            False
        )
        self.assertEqual(
            container_region.contains(Region(12, 13, "+", "chr2")),
            False
        )

    def testStrandedContains(self):
        # Forward strand
        f_region = Region(9, 15, "+", "chr1")
        self.assertEqual(
            f_region.strandedContains(Region(9, 12, "+", "chr1")),
            True
        )
        self.assertEqual(
            f_region.strandedContains(Region(9, 12, "-", "chr1")),
            False
        )
        # Reverse strand
        r_region = Region(9, 15, "-", "chr1")
        self.assertEqual(
            r_region.strandedContains(Region(9, 12, "+", "chr1")),
            False
        )
        self.assertEqual(
            r_region.strandedContains(Region(9, 12, "-", "chr1")),
            True
        )

    def testHasOverlap(self):
        region = Region(9, 15, "+", "chr1")
        self.assertEqual(
            region.hasOverlap(Region(9, 9, "+", "chr1")),
            True
        )
        self.assertEqual(
            region.hasOverlap(Region(15, 15, "+", "chr1")),
            True
        )
        self.assertEqual(
            region.hasOverlap(Region(12, 13, "+", "chr1")),
            True
        )
        self.assertEqual(
            region.hasOverlap(Region(9, 15, "+", "chr1")),
            True
        )
        self.assertEqual(
            region.hasOverlap(Region(8, 14, "+", "chr1")),
            True
        )
        self.assertEqual(
            region.hasOverlap(Region(10, 16, "+", "chr1")),
            True
        )
        self.assertEqual(
            region.hasOverlap(Region(8, 16, "+", "chr1")),
            True
        )
        self.assertEqual(
            region.hasOverlap(Region(8, 8, "+", "chr1")),
            False
        )
        self.assertEqual(
            region.hasOverlap(Region(16, 16, "+", "chr1")),
            False
        )
        self.assertEqual(
            region.hasOverlap(Region(12, 13, "+", "chr2")),
            False
        )

    def testHasStrandedOverlap(self):
        # Forward strand
        f_region = Region(9, 15, "+", "chr1")
        self.assertEqual(
            f_region.hasStrandedOverlap(Region(5, 12, "+", "chr1")),
            True
        )
        self.assertEqual(
            f_region.hasStrandedOverlap(Region(5, 12, "-", "chr1")),
            False
        )
        # Reverse strand
        r_region = Region(9, 15, "-", "chr1")
        self.assertEqual(
            r_region.hasStrandedOverlap(Region(5, 12, "+", "chr1")),
            False
        )
        self.assertEqual(
            r_region.hasStrandedOverlap(Region(5, 12, "-", "chr1")),
            True
        )

    def testGetMinDist(self):
        region = Region(9, 15, "+", "chr1")
        self.assertEqual(
            region.getMinDist(Region(14, 18, "+", "chr1")),
            0
        )
        self.assertEqual(
            region.getMinDist(Region(16, 18, "+", "chr1")),
            1
        )
        self.assertEqual(
            region.getMinDist(Region(1, 5, "+", "chr1")),
            4
        )
        with self.assertRaises(Exception):
            region.getMinDist(Region(1, 5, "+", "chr2"))

    def testIterOverlapped_1(self):
        """Case where it does not exist any overlap between subjects."""
        # Init test data
        sbjct_1 = Region(7, 10, "+", "chr1", "sbjct_1")
        sbjct_2 = Region(14, 17, "+", "chr1", "sbjct_2")
        sbjct_3 = Region(24, 29, "+", "chr1", "sbjct_3")
        subjects = RegionList([sbjct_1, sbjct_2, sbjct_3])
        queries_info = [
            {"query": Region(11, 11, "+", "chr1", "query_l1_01"), "overlapped": []},
            {"query": Region(12, 12, "+", "chr1", "query_l1_02"), "overlapped": []},
            {"query": Region(13, 13, "+", "chr1", "query_l1_03"), "overlapped": []},
            {"query": Region(14, 14, "+", "chr1", "query_l1_04"), "overlapped": [sbjct_2]},
            {"query": Region(15, 15, "+", "chr1", "query_l1_05"), "overlapped": [sbjct_2]},
            {"query": Region(16, 16, "+", "chr1", "query_l1_06"), "overlapped": [sbjct_2]},
            {"query": Region(17, 17, "+", "chr1", "query_l1_07"), "overlapped": [sbjct_2]},
            {"query": Region(18, 18, "+", "chr1", "query_l1_08"), "overlapped": []},
            {"query": Region(19, 19, "+", "chr1", "query_l1_09"), "overlapped": []},
            {"query": Region(20, 20, "+", "chr1", "query_l1_10"), "overlapped": []},
            {"query": Region(21, 21, "+", "chr1", "query_l1_11"), "overlapped": []},
            {"query": Region(22, 22, "+", "chr1", "query_l1_12"), "overlapped": []},
            {"query": Region(11, 13, "+", "chr1", "query_l3_01"), "overlapped": []},
            {"query": Region(12, 14, "+", "chr1", "query_l3_02"), "overlapped": [sbjct_2]},
            {"query": Region(13, 15, "+", "chr1", "query_l3_03"), "overlapped": [sbjct_2]},
            {"query": Region(14, 16, "+", "chr1", "query_l3_04"), "overlapped": [sbjct_2]},
            {"query": Region(15, 17, "+", "chr1", "query_l3_05"), "overlapped": [sbjct_2]},
            {"query": Region(16, 18, "+", "chr1", "query_l3_06"), "overlapped": [sbjct_2]},
            {"query": Region(17, 19, "+", "chr1", "query_l3_07"), "overlapped": [sbjct_2]},
            {"query": Region(18, 20, "+", "chr1", "query_l3_08"), "overlapped": []},
            {"query": Region(19, 21, "+", "chr1", "query_l3_09"), "overlapped": []},
            {"query": Region(20, 22, "+", "chr1", "query_l3_10"), "overlapped": []},
            {"query": Region(21, 23, "+", "chr1", "query_l3_11"), "overlapped": []},
            {"query": Region(13, 17, "+", "chr1", "query_l5_01"), "overlapped": [sbjct_2]},
            {"query": Region(15, 19, "+", "chr1", "query_l5_02"), "overlapped": [sbjct_2]},
            {"query": Region(17, 21, "+", "chr1", "query_l5_03"), "overlapped": [sbjct_2]},
            {"query": Region(13, 19, "+", "chr1", "query_l5_04"), "overlapped": [sbjct_2]},
            {"query": Region(13, 18, "+", "chr1", "query_l6_01"), "overlapped": [sbjct_2]},
            {"query": Region(14, 19, "+", "chr1", "query_l6_02"), "overlapped": [sbjct_2]},
            {"query": Region(15, 20, "+", "chr1", "query_l6_03"), "overlapped": [sbjct_2]},
            {"query": Region(16, 21, "+", "chr1", "query_l6_04"), "overlapped": [sbjct_2]},
            {"query": Region(17, 22, "+", "chr1", "query_l6_05"), "overlapped": [sbjct_2]},
            {"query": Region(18, 23, "+", "chr1", "query_l6_06"), "overlapped": []},
            {"query": Region(13, 19, "+", "chr1", "query_l7_01"), "overlapped": [sbjct_2]},
            {"query": Region(14, 20, "+", "chr1", "query_l7_02"), "overlapped": [sbjct_2]},
            {"query": Region(15, 21, "+", "chr1", "query_l7_03"), "overlapped": [sbjct_2]},
            {"query": Region(16, 22, "+", "chr1", "query_l7_04"), "overlapped": [sbjct_2]},
            {"query": Region(17, 23, "+", "chr1", "query_l7_05"), "overlapped": [sbjct_2]},
            {"query": Region(13, 20, "+", "chr1", "query_l8_01"), "overlapped": [sbjct_2]},
            {"query": Region(14, 21, "+", "chr1", "query_l8_02"), "overlapped": [sbjct_2]},
            {"query": Region(15, 22, "+", "chr1", "query_l8_03"), "overlapped": [sbjct_2]},
            {"query": Region(13, 21, "+", "chr1", "query_l9_01"), "overlapped": [sbjct_2]},
            {"query": Region(14, 22, "+", "chr1", "query_l9_02"), "overlapped": [sbjct_2]},
            {"query": Region(13, 22, "+", "chr1", "query_l10_01"), "overlapped": [sbjct_2]}
        ]
        queries_info = sorted(queries_info, key=lambda x: (x["query"].start, x["query"].end))
        # Independant evaluation
        for curr_eval in queries_info:
            obs_overlapped = [overlapped_subjects for query, overlapped_subjects in iterOverlapped([curr_eval["query"]], subjects)]
            self.assertEqual(obs_overlapped, [curr_eval["overlapped"]])
        # Grouped evaluation
        queries = [curr_info["query"] for curr_info in queries_info]
        expec_overlapped = [curr_info["overlapped"] for curr_info in queries_info]
        obs_overlapped = [overlapped_subjects for query, overlapped_subjects in iterOverlapped(queries, subjects)]
        self.assertEqual(obs_overlapped, expec_overlapped)
        # Grouped evaluation and overlapped subject starts the list of subject
        shift_subjects = subjects[1:]
        queries = [curr_info["query"] for curr_info in queries_info]
        expec_overlapped = [curr_info["overlapped"] for curr_info in queries_info]
        obs_overlapped = [overlapped_subjects for query, overlapped_subjects in iterOverlapped(queries, shift_subjects)]
        self.assertEqual(obs_overlapped, expec_overlapped)
        # Grouped evaluation and overlapped subject ends the list of subject
        pop_subjects = subjects[:-1]
        queries = [curr_info["query"] for curr_info in queries_info]
        expec_overlapped = [curr_info["overlapped"] for curr_info in queries_info]
        obs_overlapped = [overlapped_subjects for query, overlapped_subjects in iterOverlapped(queries, pop_subjects)]
        self.assertEqual(obs_overlapped, expec_overlapped)

    def testIterOverlapped_2(self):
        """Case where it exists a partial overlap between two subjects."""
        # Init test datas
        sbjct_1 = Region(7, 10, "+", "chr1", "sbjct_1")
        sbjct_2 = Region(14, 18, "+", "chr1", "sbjct_2")
        sbjct_3 = Region(16, 20, "+", "chr1", "sbjct_3")
        sbjct_4 = Region(24, 29, "+", "chr1", "sbjct_4")
        subjects = RegionList([sbjct_1, sbjct_2, sbjct_3, sbjct_4])
        queries_info = [
            {"query": Region(11, 11, "+", "chr1", "query_l1_01"), "overlapped": []},
            {"query": Region(12, 12, "+", "chr1", "query_l1_02"), "overlapped": []},
            {"query": Region(13, 13, "+", "chr1", "query_l1_03"), "overlapped": []},
            {"query": Region(14, 14, "+", "chr1", "query_l1_04"), "overlapped": [sbjct_2]},
            {"query": Region(15, 15, "+", "chr1", "query_l1_05"), "overlapped": [sbjct_2]},
            {"query": Region(16, 16, "+", "chr1", "query_l1_06"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(17, 17, "+", "chr1", "query_l1_07"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(18, 18, "+", "chr1", "query_l1_08"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(19, 19, "+", "chr1", "query_l1_09"), "overlapped": [sbjct_3]},
            {"query": Region(20, 20, "+", "chr1", "query_l1_10"), "overlapped": [sbjct_3]},
            {"query": Region(21, 21, "+", "chr1", "query_l1_11"), "overlapped": []},
            {"query": Region(22, 22, "+", "chr1", "query_l1_12"), "overlapped": []},
            {"query": Region(23, 23, "+", "chr1", "query_l1_13"), "overlapped": []},
            {"query": Region(24, 24, "+", "chr1", "query_l1_14"), "overlapped": [sbjct_4]},
            {"query": Region(11, 13, "+", "chr1", "query_l3_01"), "overlapped": []},
            {"query": Region(12, 14, "+", "chr1", "query_l3_02"), "overlapped": [sbjct_2]},
            {"query": Region(13, 15, "+", "chr1", "query_l3_03"), "overlapped": [sbjct_2]},
            {"query": Region(14, 16, "+", "chr1", "query_l3_04"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(15, 17, "+", "chr1", "query_l3_05"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(16, 18, "+", "chr1", "query_l3_06"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(17, 19, "+", "chr1", "query_l3_07"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(18, 20, "+", "chr1", "query_l3_08"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(19, 21, "+", "chr1", "query_l3_09"), "overlapped": [sbjct_3]},
            {"query": Region(20, 22, "+", "chr1", "query_l3_10"), "overlapped": [sbjct_3]},
            {"query": Region(21, 23, "+", "chr1", "query_l3_11"), "overlapped": []},
            {"query": Region(13, 17, "+", "chr1", "query_l5_01"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(15, 19, "+", "chr1", "query_l5_02"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(17, 21, "+", "chr1", "query_l5_03"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(18, 22, "+", "chr1", "query_l5_04"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(19, 23, "+", "chr1", "query_l5_05"), "overlapped": [sbjct_3]},
            {"query": Region(20, 24, "+", "chr1", "query_l5_06"), "overlapped": [sbjct_3, sbjct_4]},
            {"query": Region(13, 18, "+", "chr1", "query_l6_01"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(14, 19, "+", "chr1", "query_l6_02"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(15, 20, "+", "chr1", "query_l6_03"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(16, 21, "+", "chr1", "query_l6_04"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(17, 22, "+", "chr1", "query_l6_05"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(18, 23, "+", "chr1", "query_l6_06"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(19, 24, "+", "chr1", "query_l6_07"), "overlapped": [sbjct_3, sbjct_4]},
            {"query": Region(13, 19, "+", "chr1", "query_l7_01"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(14, 20, "+", "chr1", "query_l7_02"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(15, 21, "+", "chr1", "query_l7_03"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(16, 22, "+", "chr1", "query_l7_04"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(17, 23, "+", "chr1", "query_l7_05"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(18, 24, "+", "chr1", "query_l7_06"), "overlapped": [sbjct_2, sbjct_3, sbjct_4]},
            {"query": Region(19, 25, "+", "chr1", "query_l7_07"), "overlapped": [sbjct_3, sbjct_4]},
            {"query": Region(13, 20, "+", "chr1", "query_l8_01"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(14, 21, "+", "chr1", "query_l8_02"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(15, 22, "+", "chr1", "query_l8_03"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(13, 21, "+", "chr1", "query_l9_01"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(14, 22, "+", "chr1", "query_l9_02"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(13, 22, "+", "chr1", "query_l10_01"), "overlapped": [sbjct_2, sbjct_3]}
        ]
        queries_info = sorted(queries_info, key=lambda x: (x["query"].start, x["query"].end))
        # Independant evaluation
        for curr_eval in queries_info:
            obs_overlapped = [overlapped_subjects for query, overlapped_subjects in iterOverlapped([curr_eval["query"]], subjects)]
            self.assertEqual(obs_overlapped, [curr_eval["overlapped"]])
        # Grouped evaluation
        queries = [curr_info["query"] for curr_info in queries_info]
        expec_overlapped = [curr_info["overlapped"] for curr_info in queries_info]
        obs_overlapped = [overlapped_subjects for query, overlapped_subjects in iterOverlapped(queries, subjects)]
        self.assertEqual(obs_overlapped, expec_overlapped)
        # Grouped evaluation and overlap between subjects starts the list of subjects
        shifted_subjects = subjects[1:]
        queries = [curr_info["query"] for curr_info in queries_info]
        expec_overlapped = [curr_info["overlapped"] for curr_info in queries_info]
        obs_overlapped = [overlapped_subjects for query, overlapped_subjects in iterOverlapped(queries, shifted_subjects)]
        self.assertEqual(obs_overlapped, expec_overlapped)
        # Grouped evaluation and overlap between subjects ends the list of subjects
        poped_subjects = subjects[:-1]
        queries = [curr_info["query"] for curr_info in queries_info]
        expec_overlapped = []
        for curr_info in queries_info:
            expec_overlapped.append([elt for elt in curr_info["overlapped"] if elt != sbjct_4])
        obs_overlapped = [overlapped_subjects for query, overlapped_subjects in iterOverlapped(queries, poped_subjects)]
        self.assertEqual(obs_overlapped, expec_overlapped)

    def testIterOverlapped_3(self):
        """Case where a subject is included in another."""
        # Init test data
        sbjct_1 = Region(7, 10, "+", "chr1", "sbjct_1")
        sbjct_2 = Region(14, 20, "+", "chr1", "sbjct_2")
        sbjct_3 = Region(16, 18, "+", "chr1", "sbjct_3")
        sbjct_4 = Region(24, 29, "+", "chr1", "sbjct_4")
        subjects = RegionList([sbjct_1, sbjct_2, sbjct_3, sbjct_4])
        queries_info = [
            {"query": Region(11, 11, "+", "chr1", "query_l1_01"), "overlapped": []},
            {"query": Region(12, 12, "+", "chr1", "query_l1_02"), "overlapped": []},
            {"query": Region(13, 13, "+", "chr1", "query_l1_03"), "overlapped": []},
            {"query": Region(14, 14, "+", "chr1", "query_l1_04"), "overlapped": [sbjct_2]},
            {"query": Region(15, 15, "+", "chr1", "query_l1_05"), "overlapped": [sbjct_2]},
            {"query": Region(16, 16, "+", "chr1", "query_l1_06"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(17, 17, "+", "chr1", "query_l1_07"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(18, 18, "+", "chr1", "query_l1_08"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(19, 19, "+", "chr1", "query_l1_09"), "overlapped": [sbjct_2]},
            {"query": Region(20, 20, "+", "chr1", "query_l1_10"), "overlapped": [sbjct_2]},
            {"query": Region(21, 21, "+", "chr1", "query_l1_11"), "overlapped": []},
            {"query": Region(22, 22, "+", "chr1", "query_l1_12"), "overlapped": []},
            {"query": Region(11, 13, "+", "chr1", "query_l3_01"), "overlapped": []},
            {"query": Region(12, 14, "+", "chr1", "query_l3_02"), "overlapped": [sbjct_2]},
            {"query": Region(13, 15, "+", "chr1", "query_l3_03"), "overlapped": [sbjct_2]},
            {"query": Region(14, 16, "+", "chr1", "query_l3_04"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(15, 17, "+", "chr1", "query_l3_05"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(16, 18, "+", "chr1", "query_l3_06"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(17, 19, "+", "chr1", "query_l3_07"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(18, 20, "+", "chr1", "query_l3_08"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(19, 21, "+", "chr1", "query_l3_09"), "overlapped": [sbjct_2]},
            {"query": Region(20, 22, "+", "chr1", "query_l3_10"), "overlapped": [sbjct_2]},
            {"query": Region(21, 23, "+", "chr1", "query_l3_11"), "overlapped": []},
            {"query": Region(13, 17, "+", "chr1", "query_l5_01"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(15, 19, "+", "chr1", "query_l5_02"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(17, 21, "+", "chr1", "query_l5_03"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(18, 22, "+", "chr1", "query_l5_04"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(19, 23, "+", "chr1", "query_l5_05"), "overlapped": [sbjct_2]},
            {"query": Region(20, 24, "+", "chr1", "query_l5_06"), "overlapped": [sbjct_2, sbjct_4]},
            {"query": Region(13, 18, "+", "chr1", "query_l6_01"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(14, 19, "+", "chr1", "query_l6_02"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(15, 20, "+", "chr1", "query_l6_03"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(16, 21, "+", "chr1", "query_l6_04"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(17, 22, "+", "chr1", "query_l6_05"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(18, 23, "+", "chr1", "query_l6_06"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(19, 24, "+", "chr1", "query_l6_07"), "overlapped": [sbjct_2, sbjct_4]},
            {"query": Region(13, 19, "+", "chr1", "query_l7_01"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(14, 20, "+", "chr1", "query_l7_02"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(15, 21, "+", "chr1", "query_l7_03"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(16, 22, "+", "chr1", "query_l7_04"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(17, 23, "+", "chr1", "query_l7_05"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(18, 24, "+", "chr1", "query_l7_06"), "overlapped": [sbjct_2, sbjct_3, sbjct_4]},
            {"query": Region(19, 24, "+", "chr1", "query_l7_07"), "overlapped": [sbjct_2, sbjct_4]},
            {"query": Region(13, 20, "+", "chr1", "query_l8_01"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(14, 21, "+", "chr1", "query_l8_02"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(15, 22, "+", "chr1", "query_l8_03"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(13, 21, "+", "chr1", "query_l9_01"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(14, 22, "+", "chr1", "query_l9_02"), "overlapped": [sbjct_2, sbjct_3]},
            {"query": Region(13, 22, "+", "chr1", "query_l10_01"), "overlapped": [sbjct_2, sbjct_3]}
        ]
        queries_info = sorted(queries_info, key=lambda x: (x["query"].start, x["query"].end))
        # Independant evaluation
        for curr_eval in queries_info:
            obs_overlapped = [overlapped_subjects for query, overlapped_subjects in iterOverlapped([curr_eval["query"]], subjects)]
            self.assertEqual(obs_overlapped, [curr_eval["overlapped"]])
        # Grouped evaluation
        queries = [curr_info["query"] for curr_info in queries_info]
        expec_overlapped = [curr_info["overlapped"] for curr_info in queries_info]
        obs_overlapped = [overlapped_subjects for query, overlapped_subjects in iterOverlapped(queries, subjects)]
        self.assertEqual(obs_overlapped, expec_overlapped)
        # Grouped evaluation and inclusion between subjects starts the list of subjects
        shifted_subjects = subjects[1:]
        queries = [curr_info["query"] for curr_info in queries_info]
        expec_overlapped = [curr_info["overlapped"] for curr_info in queries_info]
        obs_overlapped = [overlapped_subjects for query, overlapped_subjects in iterOverlapped(queries, shifted_subjects)]
        self.assertEqual(obs_overlapped, expec_overlapped)
        # Grouped evaluation and inclusion between subjects ends the list of subjects
        poped_subjects = subjects[:-1]
        queries = [curr_info["query"] for curr_info in queries_info]
        expec_overlapped = []
        for curr_info in queries_info:
            expec_overlapped.append([elt for elt in curr_info["overlapped"] if elt != sbjct_4])
        obs_overlapped = [overlapped_subjects for query, overlapped_subjects in iterOverlapped(queries, poped_subjects)]
        self.assertEqual(obs_overlapped, expec_overlapped)

    def testFromStr(self):
        observed = Region.fromStr("12:1534187-1534287")
        expected = Region(1534187, 1534287, None, "12")
        self.assertEqual(str(observed), str(expected))


class TestRegionTree(unittest.TestCase):
    def testInit(self):
        exon1 = RegionTree(10, 30, "+", "chr1")
        exon2 = RegionTree(40, 70, "+", "chr1")
        exon3 = RegionTree(80, 100, "+", "chr1")
        RegionTree(children=[exon2, exon3, exon1])
        RegionTree(10, 100, "+", "chr1", children=[exon2, exon3, exon1])

    def testStart(self):
        tr1_region = RegionTree(10, 100, "+", "chr1")
        tr2_region = RegionTree(None, None, None, "chr1")
        exon1_region = RegionTree(10, 30, "+", "chr1")
        exon2_region = RegionTree(40, 70, "+", "chr1")
        exon3_region = RegionTree(80, 100, "+", "chr1")

        tr1_region.addChild(exon1_region)
        tr1_region.addChild(exon2_region)
        tr1_region.addChild(exon3_region)
        self.assertEqual(tr1_region.start, 10)

        tr2_region.addChild(exon1_region)
        tr2_region.addChild(exon2_region)
        tr2_region.addChild(exon3_region)
        self.assertEqual(tr2_region.start, 10)

    def testEnd(self):
        tr1_region = RegionTree(10, 100, "+", "chr1")
        tr2_region = RegionTree(None, None, None, "chr1")
        exon1_region = RegionTree(10, 30, "+", "chr1")
        exon2_region = RegionTree(40, 70, "+", "chr1")
        exon3_region = RegionTree(80, 100, "+", "chr1")

        tr1_region.addChild(exon1_region)
        tr1_region.addChild(exon2_region)
        tr1_region.addChild(exon3_region)
        self.assertEqual(tr1_region.end, 100)

        tr2_region.addChild(exon1_region)
        tr2_region.addChild(exon2_region)
        tr2_region.addChild(exon3_region)
        self.assertEqual(tr2_region.end, 100)

    def testStrand(self):
        # Forward strand
        tr1_region = RegionTree(10, 100, "+", "chr1")
        tr2_region = RegionTree(None, None, None, "chr1")
        exon1_region = RegionTree(10, 30, "+", "chr1")
        exon2_region = RegionTree(40, 70, "+", "chr1")
        exon3_region = RegionTree(80, 100, "+", "chr1")

        tr1_region.addChild(exon1_region)
        tr1_region.addChild(exon2_region)
        tr1_region.addChild(exon3_region)
        self.assertEqual(tr1_region.strand, "+")

        tr2_region.addChild(exon1_region)
        tr2_region.addChild(exon2_region)
        tr2_region.addChild(exon3_region)
        self.assertEqual(tr2_region.strand, "+")

        # Reverse strand
        tr1_region = RegionTree(10, 100, "-", "chr1")
        tr2_region = RegionTree(None, None, None, "chr1")
        exon1_region = RegionTree(10, 30, "-", "chr1")
        exon2_region = RegionTree(40, 70, "-", "chr1")
        exon3_region = RegionTree(80, 100, "-", "chr1")

        tr1_region.addChild(exon1_region)
        tr1_region.addChild(exon2_region)
        tr1_region.addChild(exon3_region)
        self.assertEqual(tr1_region.strand, "-")

        tr2_region.addChild(exon1_region)
        tr2_region.addChild(exon2_region)
        tr2_region.addChild(exon3_region)
        self.assertEqual(tr2_region.strand, "-")

    def testAddChild(self):
        tr1_region = RegionTree(None, None, None, "chr1")
        exon1_region = RegionTree(10, 30, "-", "chr1", "exon1")
        exon2_region = RegionTree(40, 70, "-", "chr1", "exon2")
        exon3_region = RegionTree(80, 100, "-", "chr1", "exon3")
        tr1_region.addChild(exon2_region)
        tr1_region.addChild(exon1_region)
        tr1_region.addChild(exon3_region)
        self.assertEqual(
            [str(child) for child in tr1_region.children],
            [str(exon3_region), str(exon2_region), str(exon1_region)]
        )
        self.assertEqual(
            [str(child.parent) for child in tr1_region.children],
            [str(tr1_region), str(tr1_region), str(tr1_region)]
        )
        exon4_region = RegionTree(105, 120, "+", "chr1")
        with self.assertRaises(ValueError):
            tr1_region.addChild(exon4_region)
        exon5_region = RegionTree(105, 120, "-", "chr2")
        with self.assertRaises(ValueError):
            tr1_region.addChild(exon5_region)


class TestSplittedByRef(unittest.TestCase):
    def testSplittedByRef(self):
        reg_list = RegionList([
            Region(10, 30, "-", "chr1", "region1"),
            Region(40, 70, "-", "chr1", "region2"),
            Region(80, 100, "-", "chr2", "region3")
        ])
        reg_by_chr = splittedByRef(reg_list)
        expected = ["chr1:region1", "chr1:region2", "chr2:region3"]
        observed = []
        for chrom, regions in sorted(reg_by_chr.items()):
            named_regions = []
            for curr_region in regions:
                named_regions.append("{}:{}".format(chrom, curr_region.name))
            observed.extend(named_regions)
        self.assertEqual(expected, observed)


class TestConsolidate(unittest.TestCase):
    def testConsolidate(self):
        reg_list = RegionList([
            Region(5, 9, "-", "chr1", "region1"),
            Region(10, 30, "-", "chr1", "region2"),
            Region(30, 40, "-", "chr1", "region3"),
            Region(35, 39, "-", "chr1", "region4"),
            Region(40, 70, "-", "chr1", "region5"),
            Region(71, 90, "-", "chr1", "region6"),
            Region(92, 100, "-", "chr1", "region7"),
            Region(100, 100, "+", "chr1", "region8"),
            Region(80, 100, "-", "chr2", "region9")
        ])
        # Merge overlapping
        consolidated_reg = consolidated(reg_list, False)
        expected = ["chr1:5-9[-]", "chr1:10-70[-]", "chr1:71-90[-]", "chr1:92-100[None]", "chr2:80-100[-]"]
        observed = [curr.getCoordinatesStr() for curr in consolidated_reg]
        self.assertEqual(expected, observed)
        # Merge overlapping and contiguous
        consolidated_reg = consolidated(reg_list, True)
        expected = ["chr1:5-90[-]", "chr1:92-100[None]", "chr2:80-100[-]"]
        observed = [curr.getCoordinatesStr() for curr in consolidated_reg]
        self.assertEqual(expected, observed)


########################################################################
#
# MAIN
#
########################################################################
if __name__ == "__main__":
    unittest.main()
